# -*- coding: utf-8 -*-
# Copyright (C) 2017  IRISA
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# The original code contained here was initially developed by:
#
#     Pierre Vignet.
#     IRISA
#     Dyliss team
#     IRISA Campus de Beaulieu
#     35042 RENNES Cedex, FRANCE
from __future__ import unicode_literals
from __future__ import print_function

#from pycallgraph import PyCallGraph
#from pycallgraph.output import GraphvizOutput

# Standard imports
import re
import os
from functools import partial
import sys
import itertools as it
# Multiprocessing
try:
    from concurrent.futures import ProcessPoolExecutor, as_completed
except ImportError:
    raise ImportError("No module named concurrent.futures. \
                      You can try to install 'futures' module in Python 2.7")
import multiprocessing as mp

# Custom imports
from cadbiom.models.clause_constraints.mcl.MCLAnalyser import MCLAnalyser
from cadbiom.models.clause_constraints.mcl.MCLQuery import MCLSimpleQuery
import cadbiom.commons as cm

LOGGER = cm.logger()


class ErrorRep(object):
    # Cf class CompilReporter(object):
    # gt_gui/utils/reporter.py
    def __init__(self):
        self.context = ""
        self.error = False

    def display(self, mess):
        self.error = True
        LOGGER.error(">> Context: {}; {}".format(self.context, mess))
        exit()

    def display_info(self, mess):
        LOGGER.error("-- Context: {}; {}".format(self.context, mess))
        exit()

    def set_context(self, cont):
        self.context = cont


def logical_operator(elements, operator):
    """Join elements with the given logical operator.

    :param arg1: Iterable of elements to join with a logical operator
    :param arg2: Logical operator to use 'and' or 'or'
    :return: logical_formula: str - AND/OR of the input list
    :type arg1: <list>
    :type arg2: <str>
    :rtype: <str>
    """
    assert operator in ('and', 'or')
#    print(operator + ": elements:", elements)

    return '(' + " {} ".format(operator).join(elements) + ')'


def make_logical_formula(previous_frontier_places, start_prop):
    """Make a logical formula based on previous results of MAC.

    The aim is to exclude previous solution.

    1 line: ex: "A B" => (A and B)
    another line: ex: "B C" => (B and C)
    merge all lines: (A and B) or (B and C)
    forbid all combinaisons: not((A and B) or (B and C))

    :param arg1: Set of previous frontier places (previous solutions).
    :param arg2: Original property (constraint) for the solver.
    :return: A logical formula which excludes all the previous solutions.
    :type arg1: <set>
    :type arg2: <str>
    :rtype: <str>
    """

    logical_and = partial(logical_operator, operator='and')
    logical_or = partial(logical_operator, operator='or')

    def add_start_prop(prev_frontier_places_formula):
        """Deal with start_prop if given"""

        if start_prop and prev_frontier_places_formula:
            return start_prop + ' and (' + prev_frontier_places_formula + ')'
        elif prev_frontier_places_formula:
            return prev_frontier_places_formula
        else:
            return start_prop

    cam_list = [logical_and(frontier_places)
                for frontier_places in previous_frontier_places]

    if len(cam_list) != 0:
        # Logical or between each line
        return add_start_prop('not(' + logical_or(cam_list) + ')')
    else:
        return add_start_prop('')


def main2(chart_file, cam_file, cam_step_file, cam_complete_file, cam_strong_file,
          steps, final_prop, start_prop, inv_prop, all_macs, continue_run):
    """

    :param arg1: Model file (bcx, xml, cal).
    :param arg2: File used to store Minimal Activation Condition (MAC/CAM).
    :param arg3: File used to store Minimal step numbers for each solution.
    :param arg4: File used to store MAC & trajectories.
    :param arg5: ???
    :param arg6: Maximal steps to reach the solutions.
    :param arg7: Formula: Property that the solver looks for.
    :param arg8: Formula: Original property (constraint) for the solver.
    :param arg9: Formula: ???
    :param arg10: If set to True (not default), search all macs with
        less or equal steps than previous, by limiting steps.
    :param arg11: If set to True (not default), previous macs from a previous
        run, will be reloaded.
    :type arg1: <str>
    :type arg2: <str>
    :type arg3: <str>
    :type arg4: <str>
    :type arg5: <str>
    :type arg6: <int>
    :type arg7: <str>
    :type arg8: <str>
    :type arg9: <str>
    :type arg10: <boolean>
    :type arg11: <boolean>

    .. todo: handle these QUERY PARAMETERS... from GUI program

            #    if self.possible:
            #        if len(inv_prop) == 0:
            #            inv_prop = None
            #        else :
            #            inv_prop = "not ("+inv_prop+")"
            #    else:
            #        if len(inv_prop) != 0:
            # sert pas:
            #            final_prop = "not ("+final_prop+" and "+inv_prop+")"
    """


    # Build MCLA with Error Reporter
    mcla = MCLAnalyser(ErrorRep())

    # Load model file
    detect_model_type(mcla, chart_file)(chart_file)
    if mcla.reporter.error:
        raise "Error during loading of file"


    # Frontier places asked
    if continue_run:
        # Reload previous working files
        try:
            previous_frontier_places = read_cam_file(cam_file)
            current_start_prop = make_logical_formula(previous_frontier_places,
                                                      start_prop)
            LOGGER.info(final_prop + \
                        ":: Reload previous frontier places: " + \
                        str(len(previous_frontier_places)))
        except IOError:
            LOGGER.warning(final_prop + ":: cam file not found !")
            previous_frontier_places = set()
            current_start_prop = start_prop
    else:
        # New run
        previous_frontier_places = set()
        current_start_prop = start_prop


#    i = 0
#    with PyCallGraph(output=GraphvizOutput()):
    while True:
        print("START PROP:", current_start_prop)
        ret = \
            find_mac(mcla,
                     cam_file, cam_step_file, cam_complete_file,
                     steps, final_prop, current_start_prop, inv_prop)
        # EXIT
#        i += 1
#        if i == 3:
#            return
        if ret:
            frontier_places, min_steps = ret
        else:
            # No new solution/not satisfiable
            return

        # Add theese frontier places to set of previous ones
        # (tuple is hashable)
        previous_frontier_places.add(tuple(frontier_places))
        LOGGER.debug(final_prop + ":: Prev frontier places: " + \
                     str(previous_frontier_places))

        # Compute the formula of the next start_property
        current_start_prop = make_logical_formula(previous_frontier_places,
                                                  start_prop)

        # If all_macs flag is not set (to True),
        # search allways macs with less or equal steps than previous,
        # by limiting steps.
        # If all_macs == True: The limit is always the maximal number given
        # in 'step' setting.
        if not all_macs:
            steps = min_steps

        LOGGER.debug(
            final_prop +
            ":: Next start_prop formula: {} in {} steps".format(
                current_start_prop,
                steps
            )
        )


def find_macs(mcla,
              cam_file, cam_step_file, cam_complete_file,
              steps, final_prop, start_prop, inv_prop):
    """__mac_exhaustive_search, on cherche d´abord des solutions non minimales
    (lfsol = self.__sq_dimacs_frontier_solutions(query, nb_step, 2))
    pour ensuite les élaguer en supprimer les places non essentielles à la
    satisfiabilité de la propriété
    (current_mac = self.__prune_frontier_solution(small_sol, query, nb_step)).
    Ce processus récursif est le plus "time-consuming" car nous n´avons pas
    le control sur les solutions fournies par SAT et les solutions non minimale
    sont en générales tres éloignées de la solution minimale,
    cad contiennent beaucoup plus de places.

    :return: None
    """

    # Build query
    query = MCLSimpleQuery(start_prop, inv_prop, final_prop)

    vmac_list = mcla.mac_search(query, steps)

    # If yes, in how many steps ?
    min_step   = mcla.unfolder.get_current_step()
    for next_mac_object in vmac_list:

        LOGGER.debug(final_prop + ":: Next MAC object:\n{}".format(next_mac_object))

        # Save MAC and timings
        LOGGER.debug(final_prop + ":: Save MAC and timings...")
        with open(cam_complete_file, 'a') as file:
            next_mac_object.save(file)

        # Save MAC
        next_mac = next_mac_object.activated_frontier
        LOGGER.debug(final_prop + ":: Save next MAC: {}".format(next_mac))
        with open(cam_file, 'a') as file:
            file.write('\t'.join(next_mac) + '\n')

    # Save min steps
    min_step = mcla.unfolder.get_current_step() - 1 # Magic number !
    LOGGER.debug(final_prop + ":: Save minimal steps: {}".format(min_step))
    with open(cam_step_file, 'a') as file:
        file.write(str(min_step)+'\n')


def find_mac(mcla,
             cam_file, cam_step_file, cam_complete_file,
             steps, final_prop, start_prop, inv_prop):
    """
    list, min step
    :return: None if there is no new Solution or if problem is not satisfiable.
    """
    # Build query
    query = MCLSimpleQuery(start_prop, inv_prop, final_prop)

    # Is the property reacheable ?
    reacheable = mcla.sq_is_satisfiable(query, steps)
    # If yes, in how many steps ?
    min_step   = mcla.unfolder.get_current_step()

    if reacheable and (min_step <= steps):
        LOGGER.info(
            final_prop + ":: Property {} is reacheable in {} steps".format(
                final_prop, min_step
            )
        )
    else:
        LOGGER.info(
            final_prop + ":: Property {} is NOT reacheable in {} steps".format(
                final_prop, min_step
            )
        )
        LOGGER.info(final_prop + ":: STOP the search!")
        return

    # Find next MAC
    next_mac_object = mcla.next_mac(query, min_step)
    if next_mac_object:
        LOGGER.debug(final_prop + ":: Next MAC object:\n{}".format(next_mac_object))

        # Save MAC and timings
        LOGGER.debug(final_prop + ":: Save MAC and timings...")
        with open(cam_complete_file, 'a') as file:
            next_mac_object.save(file)

        # Save MAC (in alphabetic order...)
        next_mac = sorted(next_mac_object.activated_frontier)
        LOGGER.debug(final_prop + ":: Save next MAC: {}".format(next_mac))
        with open(cam_file, 'a') as file:
            file.write(" ".join(sorted(next_mac, key=lambda s: s.lower()))+'\n')

        # Save min steps
        min_step = mcla.unfolder.get_current_step() - 1 # Magic number !
        LOGGER.debug(final_prop + ":: Save minimal steps: {}".format(min_step))
        with open(cam_step_file, 'a') as file:
            file.write(str(min_step)+'\n')

        return next_mac, min_step
    else:
        LOGGER.info(final_prop + ":: STOP the search! No more MAC.")
        return


def detect_model_type(mclanalyser, filepath):
    """Return the function to use to load the model.

    The detection is based on the file extension.

    bcx file: Build an MCLAnalyser from a .bcx file:
        build_from_chart_file()
    cal file: Build an MCLAnalyser from a .cal file of PID database
        build_from_cadlang()
    xml file: Build an MCLAnalyser from a .xml file of PID database:
        build_from_pid_file()

    :param arg1: MCLAnalyser.
    :param arg2: File that contains the model.
    :type arg1: <MCLAnalyser>
    :type arg2: <str>
    :return: The function to use to read the given file.
    :rtype: <func>
    """

    build_func = {".bcx": mclanalyser.build_from_chart_file,
                  ".cal": mclanalyser.build_from_cadlang,
                  ".xml": mclanalyser.build_from_pid_file,
                 }
    try:
        extension = re.search('^.*(\.[bcx|xml|cal]{3})', filepath).group(1)
        LOGGER.debug("Find '" + extension + "' extension: " + filepath)
    except:
        LOGGER.error("Unauthorized file: " + filepath)
        exit()

    return build_func[extension]


#def main(chart_file, cam_file, cam_step_file, cam_complete_file, cam_strong_file,
#         steps, final_prop, start_prop, inv_prop, all_macs):
#
#    LOGGER.debug("Params: start: {}, inv: {}, final: {}".format(start_prop,
#                                                                inv_prop,
#                                                                final_prop))
#
#    cam_p = None
#    # forbid previous cam
#    try :
#        cam_p = camFile2notOr(cam_file)
#        print("cam file:", cam_p)
#    except :
#        print('error in camFile2notOr')
#
#    if start_prop and cam_p :
#        start_prop += ' and ('+cam_p+')'
#    elif cam_p :
#        start_prop = cam_p
#
#    # BUILD MCLA
#    error_reporter = ErrorRep()
#    mcla = MCLAnalyser(error_reporter)
#    mcla.build_from_chart_file(chart_file)
#
#    if error_reporter.error:
#        #cancel_warn(error_reporter.memory)
#        #error_reporter.reset()
#        #return
#        print("ERROR:: " + error_reporter.error)
#        raise
##
#    # BUILD QUERY
#    query = MCLSimpleQuery(start_prop, inv_prop, final_prop)
#
#    # Frontier init:
#    # on_yn function:
#    # reach = self.mcla.sq_is_satisfiable(query, max_step)
#
#    # Show solutions:
#    #on_cond function:
##    lsol = mcla.sq_frontier_solutions(query, step, 10)
##    print("DEBUG:: " + str(lsol))
##    print("DEBUG:: ", len( lsol))
#
#    # minimal activation conditions
#    # on_cam function:
##    cam_list = self.mcla.mac_search(query, self.max_step)
##            if len(cam_list)==0 :
##                ok_warn("The solver returns an empty list" +
##                        "\n"+" you should refine your query")
#
#    # OPTIMIZE STEP RESEARCH
#    if os.path.isfile(cam_step_file):
#        min_step = int(get_last_line(cam_step_file))
#        print("min_step opti:", min_step)
#        query.set_steps_before_reach(min_step-1)
#
#
#    reacheable = mcla.sq_is_satisfiable(query, steps) #important step
#    print("reacheable:", reacheable)
#    min_step = mcla.unfolder.get_current_step()
#    print("min_step:", min_step)
#    # Set max step authorized
#    query.set_steps_before_reach(min_step-1)
#
#    # FIND NEXT MAC
#    next_mac_object = mcla.next_mac(query, min_step) #important step
#    print("next_mac_object:", next_mac_object)
#    if next_mac_object:
#
#        # SAVE MAC AND TIMING
#        with open(cam_complete_file, 'a') as file:
#            next_mac_object.save(file)
#
#        # SAVE MAC
#        next_mac = next_mac_object.activated_frontier
#        print("save next mac:", next_mac)
#        write_list(next_mac, cam_file)
#
#        # SAVE STEP
#        min_step = mcla.unfolder.get_current_step()
#        print("save min step:", min_step)
#        with open(cam_step_file, 'a') as file:
#            file.write(str(min_step)+'\n')
#
#        return 0
#    else:
#        print("stop")
#        return 1


def compute_combinations(final_properties):
    """Return all combinations of final properties.

    .. note:: (in case of input_file and combinations set).

    :param: List of final properties.
    :type: <list>
    :return: List of str. Each str is a combination of final_properties linked
        by a logical 'and'.
        Ex: ('TGFB1', 'COL1A1'), ('TGFB1', 'decorin')
            gives: [u'TGFB1 and COL1A1', u'TGFB1 and decorin']

    :rtype: <list <str>>
    """

    final_properties = set(final_properties)

    def get_formula(data):
        """Include all elements in the combinations and exclude explicitely,
        all other elements.
        """

        negated_places = ' and not '.join(final_properties - data)
        return "{}{}{}".format(
                    ' and '.join(data),
                    ' and not ' if negated_places != '' else '',
                    negated_places
                )


    all_combinations = list()
    for i in range(1, len(final_properties) + 1):

        all_combinations.append(
            {get_formula(set(comb)) for comb in it.combinations(final_properties, i)}
        )

    # Unpack combinations
    all_combinations = [comb for comb in it.chain(*all_combinations)]
    LOGGER.debug("Combinations: {}, Length: {}".format(
            str(all_combinations),
            len(all_combinations)
        )
    )
    LOGGER.info("Number of computed combinations: " + \
                str(len(all_combinations)))

    return all_combinations


def launch_researchs(params):
    """Parse arguments and launch Cadbiom search of MACs
    (Minimal Activation Conditions).

    - If there is no input file, there will be only on process.
    - If an input file is given, there will be 1 process per line (per logical
    formula on each line).
    """

#    params = args_to_param(args)

    # No input file
    if params['final_prop']:
        compute_macs(params)

    else:
        # Multiple properties in input file
        # => multiprocessing: 1 process for each property

        with open(params['input_file'], 'r') as fd:
            g = (line.rstrip('\n') for line in fd)

            final_properties = [prop for prop in g if prop != '']

        if params['combinations']:
            # If input_file is set, we can compute all combinations of
            # final_properties. default: False
            final_properties = compute_combinations(final_properties)

        LOGGER.debug("Final properties: " + str(final_properties))

        # Output combinations of final_properties
#        with open(params['input_file'] + '_combinations.txt', 'w') as fd:
#            fd.write('\n'.join(final_properties) + '\n')
#
#        g = (elem for elem in final_properties)
#        for i in range(1, len(final_properties) + 1):
#            with open(params['input_file'][:-4] + '_combinations' + str(i) + '.txt', 'w') as fd:
#                try:
#                    fd.write(next(g) + '\n')
#                    fd.write(next(g) + '\n')
#                except StopIteration:
#                    break
#
#        exit()

        def update_params(prop):

            d = params.copy()
            d['final_prop'] = prop
            return d

        # Fix number of processes
        # PS: the new solver is optimized for 8 threads
#        nb_cpu_required = mp.cpu_count() / 8
#        nb_cpu_required = 1 if nb_cpu_required == 0 else nb_cpu_required

        with ProcessPoolExecutor(max_workers=mp.cpu_count()) as e:

            futures_and_output = {e.submit(compute_macs,
                                           update_params(job_property)
                                          ):job_property \
                                for job_property in final_properties} # Job name

        nb_errors = 0
        nb_done = 0
        for future in as_completed(futures_and_output):

            job_name = futures_and_output[future]

            # On affiche les résultats si les futures en contiennent.
            # Si elles contiennent une exception, on affiche l'exception.
            if future.exception() is not None:
                LOGGER.error("{} generated an exception: \n{}".format(
                    job_name,
                    future.exception())
                )
                nb_errors += 1
            else:
                # The end
                LOGGER.info("{}... \t\t[Done]".format(job_name))
                nb_done += 1

        LOGGER.info("Ending: {} errors, {} done\nbye.".format(
            nb_errors,
            nb_done)
        )


def compute_macs(params):
    """Launch Cadbiom search of MACs (Minimal Activation Conditions).

    This function is called 1 or multiple times according to the necessity
    to use multiprocessing (Cf launch_researchs()).

    .. note:: Previous result files will be deleted.

    """

    # Limit recursion
    sys.setrecursionlimit(10000)

    # QUERY PARAMETERS
    # Todo beware with type of separators in path..
    MODEL_NAME = params['chart_file'].split('/')[-1][:-4]

    # FILES
    # Add trailing '/' if not present
    output = params['output'] if params['output'][-1] == '/' \
                else params['output'] + '/'
    CAM_FILE_PREFIX   = output + MODEL_NAME + \
                        '_' + params['final_prop'] + '_cam'
    # cam_file
    cam_file          = CAM_FILE_PREFIX + ".txt"
    # cam_step_file
    cam_step_file     = CAM_FILE_PREFIX + "_step.txt"
    # cam_complete_file
    cam_complete_file =  CAM_FILE_PREFIX + "_complete.txt"
    # cam_strong_file
    cam_strong_file   = CAM_FILE_PREFIX + "_strongA.txt"

    # Reset files
    def remove_file(file):
        try:
            os.remove(file)
        except:
            pass

    if not params['continue']:
        # Reset previous working files
        # PS: the reload is done in main2() function
        remove_file(cam_file)
        remove_file(cam_step_file)
        remove_file(cam_complete_file)
        remove_file(cam_strong_file)

    # MAC research
    main2(
        params['chart_file'],   # chart_file
        cam_file,               # cam_file
        cam_step_file,          # cam_step_file
        cam_complete_file,      # cam_complete_file
        cam_strong_file,        # cam_strong_file
        params['steps'], params['final_prop'],
        params['start_prop'], params['inv_prop'],
        params['all_macs'], params['continue'],
    )

def read_cam_file(file):
    """Return a list a fontier places already found in cam file

    .. note:: use make_logical_formula() to get the new start_prop of the run.

    :param: Mac file of a previous run
    :type: <str>
    :return: A set a frontier places.
    :rtype: <set>
    """

    with open(file, 'r') as fd:
        return {tuple(line.rstrip('\n').split(' ')) for line in fd}





# -*- coding: utf-8 -*-
# Copyright (C) 2017  IRISA
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# The original code contained here was initially developed by:
#
#     Pierre Vignet.
#     IRISA
#     Dyliss team
#     IRISA Campus de Beaulieu
#     35042 RENNES Cedex, FRANCE
"""This module provides some functions to do some analyzes on the output
files of Cadbiom.
"""
from __future__ import unicode_literals
from __future__ import print_function

import os
import glob
from collections import Counter
from collections import defaultdict
import csv


## Handle output files #########################################################

def get_solutions(file_descriptor):
    """Generator of solution lines and corresponding stripped lines.

    .. note: Do not return events ! Just sets of frontier places.

    :param: Opened file.
    :type: <file>
    :return: Line (without '\n') and stripped line (with '\' replaced by ' '
        (except for final '\t')).
    :rtype: <tuple <str>, <str>>
    """

    for line in file_descriptor:
        # Remove possible \t separator from first line (frontier solution)
        line = line.rstrip('\n')
        stripped_line = line.rstrip('\t').replace('\t', ' ')

        # Next Line if empty
        if stripped_line == '':
            continue

        # Remove events or other lines
        if stripped_line[0] not in ('%', '=', ' '):
            # print(stripped_line)
            # Sort in lower case, remove ' ' empty elements
            yield line, stripped_line

## Sort functions ##############################################################

def sort_solutions_in_file(filepath):
    """Sort all solutions in the given file in alphabetical order.

    .. warning:: The file is modified in place.

    :param: Filepath to be opened and in which solutions will be sorted.
    :arg: <str>
    """

    solutions = dict()

    with open(filepath, 'r+') as fd:

        # Get old line as key and ordered line as value
        for line, stripped_line in get_solutions(fd):
            # Sort in lower case, remove ' ' empty elements
            solutions[line] = \
                " ".join(sorted([place for place in stripped_line.split(' ')
                                 if place != ' '], key=lambda s: s.lower()))

        # Rewind the whole file
        fd.seek(0)

        # Load all the content
        file_text = fd.read()

        # Replace old sols with the new ones
        for original_sol, sorted_sol in solutions.items():
            file_text = file_text.replace(original_sol, sorted_sol)

        # Rewind the whole file
        fd.seek(0)

        # Write all text in the current opened file
        fd.write(file_text)


def sort_solutions(path):
    """Entry point for sorting solutions.

    Parse a solution file and sort all frontier places in alphabetical order.

    :param: Filepath or directory path containing Cadbiom solutions.
    :type: <str>
    """

    # Check valid input file/directory
    assert os.path.isfile(path) or os.path.isdir(path)

    if os.path.isdir(path):
        # Recursive search of *cam* files
        # (cam.txt, cam_complete.txt, cam_step.txt)
        path = path if path[-1] == '/' else path + '/'
        [sort_solutions_in_file(file) for file in glob.glob(path + '*cam*')]
    else:
        sort_solutions_in_file(path)

################################################################################



# -*- coding: utf-8 -*-
"""Unit tests for Python API"""

from __future__ import unicode_literals
from __future__ import print_function


# Standard imports
import pytest

@pytest.fixture()
def feed_logical_formula():

    frontier_places = {tuple(('A', 'B')), tuple(('AB', 'CD'))}

    return frontier_places

def test_make_logical_formula(feed_logical_formula):

    from cadbiom_cmd.solution_search import make_logical_formula

    found = make_logical_formula(feed_logical_formula, None)
    assert found == "not(((AB and CD) or (A and B)))"


    found = make_logical_formula(feed_logical_formula, 'START_PROP')
    assert found == "START_PROP and (not(((AB and CD) or (A and B))))"



# -*- coding: utf-8 -*-
# Copyright (C) 2017  IRISA
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# The original code contained here was initially developed by:
#
#     Pierre Vignet.
#     IRISA
#     Dyliss team
#     IRISA Campus de Beaulieu
#     35042 RENNES Cedex, FRANCE
from __future__ import unicode_literals
from __future__ import print_function

# Standard imports
import argparse
import os

# Custom imports
import cadbiom.commons as cm

LOGGER = cm.logger()

def launch_researchs(args):
    """Parse arguments and launch Cadbiom search of MACs
    (Minimal Activation Conditions).
    """

    # Module import
    import solution_search
    params = args_to_param(args)
    solution_search.launch_researchs(params) # !


def launch_sort(args):
    """Parse a solution file or a directory of solution files,
    and sort all frontier places in alphabetical order.
    """

    # Module import
    import solution_sort
    params = args_to_param(args)
    solution_sort.sort_solutions(params['sol_file'])


def parse_trajectories(args):
    """Parse a complete solution file and make a representation of trajectories.
    """

    # Module import
    import solution_repr
    params = args_to_param(args)
    params['output'] = params['output'] if params['output'][-1] == '/' \
                        else params['output'] + '/'
    solution_repr.parse_trajectories_main(
        params['output'],
        params['chart_file'],
        params['sol_file']
    )

def sol_digging(args):
    """Convert all events for all solutions in a complete MAC file
    and write them in a separate file in the json format.

    This is a function to quickly search all transition attributes involved
    in a solution.
    """

    # Module import
    import solution_repr
    params = args_to_param(args)
    params['output'] = params['output'] if params['output'][-1] == '/' \
                        else params['output'] + '/'
    solution_repr.sol_digging_main(
        params['output'],
        params['chart_file'],
        params['sol_file'],
        conditions=not params['no_conditions'], # Inv the param...
    )

def model_comp(args):
    """Model consistency checking.

    Check if the 2 given models have the same topology,
    nodes & edges attributes/roles.
    """

    # Module import
    import solution_repr
    params = args_to_param(args)
    params['output'] = params['output'] if params['output'][-1] == '/' \
                        else params['output'] + '/'
    solution_repr.graph_isomorph_test(
        params['model_file_1'],
        params['model_file_2'],
        params['output'],
        params['graphs'],
        params['json'],
    )


def model_infos(args):
    """Model informations.

    Get number of nodes, edges, centralities (degree, closeness, betweenness).
    """

    # Module import
    import solution_repr
    params = args_to_param(args)
    params['output'] = params['output'] if params['output'][-1] == '/' \
                        else params['output'] + '/'
    solution_repr.graph_infos(
        params['model_file'],
        params['output'],
        params['graph'],
        params['json'],
        params['advanced'],
    )


def merge_cams(args):
    """Merge solutions to a csv file."""

    # Module import
    import solution_merge
    params = args_to_param(args)
    solution_merge.merge_cams_to_csv(params['solutions_directory'],
                                     params['output'])


def args_to_param(args):
    """Return argparse namespace as a dict {variable name: value}"""
    return {k: v for k, v in vars(args).items() if k != 'func'}


class ReadableFile(argparse.Action):
    """
    http://stackoverflow.com/questions/11415570/directory-path-types-with-argparse
    """

    def __call__(self, parser, namespace, values, option_string=None):
        prospective_file = values

        if not os.path.isfile(prospective_file):
            raise argparse.ArgumentTypeError(
                "readable_file:{0} is not a valid path".format(
                    prospective_file)
                )

        if os.access(prospective_file, os.R_OK):
            setattr(namespace, self.dest, prospective_file)
        else:
            raise argparse.ArgumentTypeError(
                "readable_file:{0} is not a readable file".format(
                    prospective_file)
                )


class ReadableDir(argparse.Action):
    """
    http://stackoverflow.com/questions/11415570/directory-path-types-with-argparse
    """

    def __call__(self, parser, namespace, values, option_string=None):
        prospective_dir = values

        if not os.path.isdir(prospective_dir):
            raise argparse.ArgumentTypeError(
                "readable_dir:{0} is not a valid path".format(
                    prospective_dir)
                )

        if os.access(prospective_dir, os.R_OK):
            setattr(namespace, self.dest, prospective_dir)
        else:
            raise argparse.ArgumentTypeError(
                "readable_dir:{0} is not a readable dir".format(
                    prospective_dir)
                )


def main():
    """Argument parser"""

    # parser configuration
    parser = argparse.ArgumentParser(description=__doc__)
    # Default log level: debug
    parser.add_argument('-vv', '--verbose', nargs='?', default='info')
    # Subparsers
    subparsers = parser.add_subparsers(title='subcommands')

    # PS: nargs='?' = optional
    # subparser: Compute macs
    #    steps      = 10
    #    final_prop = "P"
    #    start_prop = None
    #    inv_prop   = None
    parser_input_file = subparsers.add_parser(
        'compute_macs',
        help=launch_researchs.__doc__,
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    parser_input_file.add_argument('chart_file')
    # Get final_property alone OR an input_file containing multiple properties
    group = parser_input_file.add_mutually_exclusive_group()
    group.add_argument('final_prop', nargs='?')
    group.add_argument('--input_file', action=ReadableFile, nargs='?',
        help="Without input file, there will be only one process. " + \
        "While there will be 1 process per line (per logical formula " + \
        "on each line)")
    parser_input_file.add_argument('--output', action=ReadableDir, nargs='?',
        default='result/', help="Output directory.")

    # default: False
    parser_input_file.add_argument('--combinations', action='store_true',
        help="If input_file is set, we can compute all combinations of " + \
        "given elements on each line")
    parser_input_file.add_argument('--steps', type=int, nargs='?', default=10,
        help="Maximum of allowed steps to find macs")
    # https://docs.python.org/dev/library/argparse.html#action
    # all_macs to False by default
    parser_input_file.add_argument('--all_macs', action='store_true',
        help="Solver will try to search all macs with 0 to the maximum of " + \
        "allowed steps.")
    # continue to False by default
    parser_input_file.add_argument('--continue', action='store_true',
        help="Resume previous computations; if there is a mac file from " + \
        "a previous work, last frontier places will be reloaded.")
    parser_input_file.add_argument('--start_prop', nargs='?', default=None)
    parser_input_file.add_argument('--inv_prop', nargs='?', default=None)

    parser_input_file.set_defaults(func=launch_researchs)


    # subparser: Sort solutions in alphabetical order in place
    # Solution file (complete or not)
    parser_solutions_sort = subparsers.add_parser('sort_solutions',
                                                  help=launch_sort.__doc__)
    parser_solutions_sort.add_argument('sol_file',
        help="Solution file or directory with solution files " + \
        "(output of 'compute_macs' command).")
    parser_solutions_sort.set_defaults(func=launch_sort)


    # subparser: Representation of the trajectories of MACs in a complete file.
    # Model file (xml : cadbiom language)
    # Solution file (cam_complete)
    parser_trajectories = subparsers.add_parser(
        'parse_trajectories',
        help=parse_trajectories.__doc__,
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser_trajectories.add_argument('chart_file',
        help="bcx model file.")
    parser_trajectories.add_argument('sol_file',
        help="Complete solution file (output of 'compute_macs' command).")
    parser_trajectories.add_argument('--output', action=ReadableDir,
        nargs='?', default='graphs/',
        help="Output directory for graphml files.")
    parser_trajectories.set_defaults(func=parse_trajectories)


    # subparser: Decompilation of trajectories of MACs in a complete file/dir.
    # Model file (xml : cadbiom language)
    # Solution file (cam_complete)
    parser_sol_digging = subparsers.add_parser(
        'sol_digging',
        help=sol_digging.__doc__,
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser_sol_digging.add_argument('chart_file',
        help="bcx model file.")
    parser_sol_digging.add_argument('sol_file',
        help="Complete solution file or directory with solution files " + \
       "(output of 'compute_macs' command).")
    parser_sol_digging.add_argument('--output', action=ReadableDir,
        nargs='?', default='decompiled_solutions/',
        help="Output directory for decompiled solutions files.")
    parser_sol_digging.add_argument('--no_conditions', action='store_true',
        help="Don't export conditions of transitions. This allows " + \
        "to have only places that are used inside trajectories; " + \
        "thus, inhibitors nodes are not present in the json file")
    parser_sol_digging.set_defaults(func=sol_digging)


    # subparser: Merge solutions to a csv file
    # Solution file (cam)
    # Output (csv)
    parser_merge_cams = subparsers.add_parser(
        'merge_cams',
        help=merge_cams.__doc__,
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    parser_merge_cams.add_argument('solutions_directory', nargs='?',
        default='result/')
    parser_merge_cams.add_argument('--output', nargs='?',
        default='result/merged_cams.csv',
        help="CSV file: <Final property formula>;<cam>")
    parser_merge_cams.set_defaults(func=merge_cams)


    # subparser: Model comparison
    # 2 models
    parser_model_comparison = subparsers.add_parser(
        'model_comp',
        help=model_comp.__doc__,
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    parser_model_comparison.add_argument('model_file_1',
        help="bcx model file.")
    parser_model_comparison.add_argument('model_file_2',
        help="bcx model file.")
    # Export graphs for the 2 models; default: false
    parser_model_comparison.add_argument('--graphs', action='store_true',
        help="Create 2 graphml files from the given models.")
    parser_model_comparison.add_argument('--json', action='store_true',
        help="Create a summary dumped into a json file.")
    parser_model_comparison.add_argument('--output', action=ReadableDir,
        nargs='?', default='graphs/',
        help="Directory for created graphs files.")
    parser_model_comparison.set_defaults(func=model_comp)


    # subparser: Model infos
    # 1 model
    parser_model_infos = subparsers.add_parser(
        'model_infos',
        help=model_infos.__doc__,
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    parser_model_infos.add_argument('model_file')
    # Export graphs for the 2 models; default: false
    parser_model_infos.add_argument('--graph', action='store_true',
        help="Create a graphml file from the model.")
    parser_model_infos.add_argument('--json', action='store_true',
        help="Create a summary dumped into a json file.")
    parser_model_infos.add_argument('--advanced', action='store_true',
        help="Get centralities (degree, closeness, betweenness).")
    parser_model_infos.add_argument('--output', action=ReadableDir,
        nargs='?', default='graphs/',
        help="Directory for created graphs files.")
    parser_model_infos.set_defaults(func=model_infos)


    # get program args and launch associated command
    args = parser.parse_args()

    # Set log level
    cm.log_level(vars(args)['verbose'])

    # launch associated command
    args.func(args)



# -*- coding: utf-8 -*-
"""Unit tests for solver results"""

#from __future__ import unicode_literals
from __future__ import print_function

# Standard imports
import pytest
import tempfile
import os

# Custom imports
from cadbiom_cmd import solution_search


@pytest.fixture()
def feed_output():

    return ['Ax Bx\n'], \
           ['Ax Bx\n', '% h2 h00\n', '% h3\n', '% h0 h1\n', '% hlast\n'], \
           ['4\n']


def test_minigraph1(feed_output):
    """Test the obtention of frontier places & timings on Minigraph 1

    Output must be:

        camp:
            Ax   Bx
        cam complete:
            Ax Bx
            % h2 h00
            % h3
            % h0 h1
            % hlast
        cam step:
            4

    .. note:: Order of trajectories may change. (h2 h00, h3 vs h2, h3 h00)

    .. note:: This test is the equivalent of the following command line:

        cadbiom_cmd -vv debug compute_macs mini_test_publi.bcx Px --steps 10

    .. note:: The test directory is the temporary sytem folder.
    """

    # Create the file model in /tmp/
    # Note: prevent the deletion of the file after the close() call
    fd_model = tempfile.NamedTemporaryFile(suffix='.bcx', delete=False)
    fd_model.write(
        """<model xmlns="http://cadbiom" name="bicluster">
    <CSimpleNode name="Ax" xloc="0.156748911466" yloc="0.673568818514"/>
    <CSimpleNode name="n1" xloc="0.291727140784" yloc="0.673568818514"/>
    <CSimpleNode name="C1" xloc="0.404208998548" yloc="0.618757612668"/>
    <CSimpleNode name="n2" xloc="0.405660377359" yloc="0.722289890377"/>
    <CSimpleNode name="Px" xloc="0.577083086261" yloc="0.722140333368"/>
    <CSimpleNode name="Bx" xloc="0.157474600871" yloc="0.83922046285"/>
    <CSimpleNode name="n3" xloc="0.290570367401" yloc="0.803822907245"/>
    <CSimpleNode name="n4" xloc="0.292715433748" yloc="0.906099768906"/>
    <CSimpleNode name="C2" xloc="0.409124108889" yloc="0.803710739487"/>
    <CSimpleNode name="C3" xloc="0.408336298212" yloc="0.906099768906"/>
    <transition name="" ori="Ax" ext="n1" event="h00" condition="" action="" fact_ids="[]"/>
    <transition name="" ori="n1" ext="C1" event="h0" condition="" action="" fact_ids="[]"/>
    <transition name="" ori="n1" ext="n2" event="h1" condition="C2" action="" fact_ids="[]"/>
    <transition name="" ori="n2" ext="Px" event="hlast" condition="C1 and not C3" action="" fact_ids="[]"/>
    <transition name="" ori="Bx" ext="n3" event="h2" condition="" action="" fact_ids="[]"/>
    <transition name="" ori="Bx" ext="n4" event="h4" condition="" action="" fact_ids="[]"/>
    <transition name="" ori="n3" ext="C2" event="h3" condition="" action="" fact_ids="[]"/>
    <transition name="" ori="n4" ext="C3" event="h5" condition="" action="" fact_ids="[]"/>
    </model>"""
    )
    fd_model.close()

    # Build params
    # See the docstring for the normal command line in shell context
    params = {
        'all_macs': False,
        'chart_file': fd_model.name, # Filename + path
        'combinations': False,
        'continue': False,
        'final_prop': 'Px',
        'input_file': None,
        'inv_prop': None,
        'output': tempfile.gettempdir() + '/',
        'start_prop': None,
        'steps': 10,
        'verbose': 'debug',
    }

    # Launch the search of the minimal accessibility condition
    solution_search.launch_researchs(params)


    with open(fd_model.name[:-4] + "_Px_cam.txt", 'r') as file:
        found = [line for line in file]
        assert found == feed_output[0]

    with open(fd_model.name[:-4] + "_Px_cam_complete.txt", 'r') as file:
        found = [line for line in file]
        assert found == feed_output[1]

    with open(fd_model.name[:-4] + "_Px_cam_step.txt", 'r') as file:
        found = [line for line in file]
        assert found == feed_output[2]

    # Delete temp files
    os.remove(fd_model.name)
    os.remove(fd_model.name[:-4] + "_Px_cam.txt")
    os.remove(fd_model.name[:-4] + "_Px_cam_complete.txt")
    os.remove(fd_model.name[:-4] + "_Px_cam_step.txt")




# -*- coding: utf-8 -*-
# Copyright (C) 2017  IRISA
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# The original code contained here was initially developed by:
#
#     Pierre Vignet.
#     IRISA
#     Dyliss team
#     IRISA Campus de Beaulieu
#     35042 RENNES Cedex, FRANCE
from __future__ import unicode_literals
from __future__ import print_function

# Standard imports
import datetime as dt
from collections import defaultdict
import networkx as nx
import itertools as it
import re
import json
import os
import glob
from logging import DEBUG
# Remove matplotlib dependency
# It is used on demand during the drawing of a graph
try:
    import matplotlib.pyplot as plt
except ImportError:
    pass

# Library imports
from cadbiom.models.guard_transitions.translators.chart_xml \
    import MakeModelFromXmlFile
from cadbiom.models.biosignal.translators.gt_visitors import compile_cond, compile_event
from cadbiom.models.guard_transitions.analyser.ana_visitors import TableVisitor
from cadbiom.models.biosignal.sig_expr import *
from cadbiom.models.guard_transitions.analyser.ana_visitors import SigExpIdCollectVisitor

import cadbiom.commons as cm

LOGGER = cm.logger()

"""
Bx  Ax
% h2 h00
% h3
% h0 h1
% hlast
Bx  Ax
% h2
% h3 h00
% h0 h1
%
% hlast
Bx  Ax
% h2
% h3 h00
% h0 h1
% hlast
%
%
Bx  Ax
% h2 h00
% h3
% h0 h1
% hlast
%
%
%
"""

class Reporter(object):
    """Error reporter.

    .. note:: Link the lexer to the model allows to avoid error in Reporter
        like:  "-> dec -> Undeclared event or state"
        In practice this is time consuming and useless for what we want to do.
        See parse_condition()
    """

    def __init__(self):
        self.error = False
        self.mess = ""
        pass

    def display(self, e):
        self.error = True
        if "Undeclared event or state" not in e:
            LOGGER.debug("\t" + self.mess + " -> " + e)


def load_solutions(file):
    """Open a file with many solution/MACs.

    :param: File name
    :type: <str>
    :return: A tuple of "frontier places" and a list of events in each step.
        ("Bx Ax", [[u'h2', u'h00'], [u'h3'], [u'h0', u'h1'], [u'hlast']])
    :rtype: <tuple <str>, <list>>
    """

    sol_steps = defaultdict(list)
    sol = ""
    with open(file, 'r') as fd:
        for line in fd:
            LOGGER.debug("Load_solutions :: line: " + line)
            # Remove possible \t separator from first line (frontier solution)
            line = line.rstrip('\n').rstrip('\t').replace('\t', ' ')
            # TODO: remove last space ' ' => beware, may be informative...
            # (start transitions have no event names: ori="__start__0")
            line = line.rstrip(' ')
            if line == '' or line[0] == '=':
                # Blank line
                # Skip blank lines and solution separator in some files (=====)
                continue
            elif line[0] != '%':
                if sol == line:
                    # TODO: why this condition ?
                    # => multiple lines with the same solution ?
                    # Same frontier places
                    yield sol, sol_steps[sol]

                    # reinit sol
                    sol_steps[sol] = list()
                    continue
                elif sol == '':
                    # First sol
                    sol = line
                else:
                    # Yield previous sol
                    yield sol, sol_steps[sol]
                    sol = line

            elif line[0] == '%':
                # Remove step with only "% "
                step = line.lstrip('% ')

                if step != '':
                    sol_steps[sol].append(step.split(' '))

        # Yield last sol
        yield sol, sol_steps[sol]


def get_transitions(file, all_places=False):
    """Get all transitions in a file model (bcx format).

    :param arg1: Model in bcx format.
    :param arg2: Ask to return all the places of the model.
        (useful for get_frontier_places() that compute frontier places)
    :type arg1: <str>
    :type arg2: <bool>
    :return: A dictionnary of events as keys, and transitions as values.
        Since many transitions can define an event, values are lists.
        Each transition is a tuple with: origin node, final node, attributes
        like label and condition.
        {u'h00': [('Ax', 'n1', {u'label': u'h00[]'}),]
    :rtype: <dict <list <tuple <str>, <str>, <dict <str>: <str>>>>
        If all_places is True, this func returns also a list of all places in
        the model.
    """

    parser = MakeModelFromXmlFile(file)

#    print(dir(parser))
#    print(type(parser))
    #<type 'instance'>
    #['__doc__', '__init__', '__module__', 'get_model', 'handler', 'model', 'parser'

    g = (trans for transition in parser.handler.top_pile.transitions
         for trans in transition)

    transitions = defaultdict(list)

    for trans in g:

        # Get the names of clocks
        # Some event have many clocks (like _h_2755) for the same
        # ori/ext entities, so we have to extract them and their respective
        # conditions
        if trans.event == '':
            # null event without clock => StartNodes
            # These nodes are used to resolve the problem of
            # Strongly Connected Components (inactivated cycles in the graph)
            # The nodes
            # Avoids having SigConstExpr as event type in parse_event()
            # I create a transition (SCC-__start__?),
            # and a node (__start__?) for this case.
            trans.event = 'SCC-' + trans.ori.name
            events = {trans.event: trans.condition}
        elif re.match('_h[\w]+', trans.event):
            # 1 event (with 1 clock)
            events = {trans.event: trans.condition}
        else:
            # Many events (with many clocks with condition(s))
            events = parse_event(trans.event)

        for event, condition in events.iteritems():
            # LOGGER.debug("NEW trans", event)

            # Handle multiple transitions for 1 event
            transitions[event].append(
                (
                    trans.ori.name, trans.ext.name,
                    {
                        'label': event, #+ '[' + trans.condition + ']',
                        'condition': condition,
                    }
                )
            )

    LOGGER.info("{} transitions loaded".format(len(transitions)))
    # Return a dict instead of defaultdict to avoid later confusions
    #(masked errors) by searching a transition that was not in the model...

    assert len(transitions) != 0, "No transitions found in the model ! " \
        "Please check the names of events (_h_xxx)"

    if all_places:
        # Return all nodes
        return dict(transitions), parser.handler.node_dict.keys()

    return dict(transitions)


def get_frontier_places(transitions, all_places):
    """Return frontier places of a model (deducted from its transitions and
    from all places of the model).

    .. note:: why we use all_places from the model instead of
        (input_places - output_places) to get frontier places ?
        Because some nodes are only in conditions and not in transitions.
        If we don't do that, these nodes are missing when we compute
        valid paths from conditions.

    :param arg1: Model's transitions.
        {u'h00': [('Ax', 'n1', {u'label': u'h00[]'}),]
    :type arg1: <dict>
        keys: names of events
        values: list of transitions as tuples (with in/output, and label).
    :return: Set of frontier places.
    :rtype: <set>
    """

    # Get transitions in events
    g = tuple(trans for event in transitions.values() for trans in event)

    # Get input nodes & output nodes
#    input_places = {trans[0] for trans in g}
    output_places = {trans[1] for trans in g}

    # Get all places that are not in transitions in the "output" place
    return set(all_places) - output_places


def rec(tree, inhibitors_nodes):
    """

    tree = ('H', 'v', (
        ('F', 'v', 'G'),
        '^',
        (
            ('A', 'v', 'B'),
            '^',
            ('C', 'v', ('D', '^', 'E'))
        )
    ))
    """

#    print("TREE", tree, type(tree), dir(tree))

    if isinstance(tree, str):  # terminal node
        path = [tree]
        solutions = [path]
        return solutions
    if isinstance(tree, SigNotExpr):
        LOGGER.debug("NOT OPERAND: {}, {}".\
            format(
                tree.operand,
                type(tree.operand)
            )
        )
        try:
            current_inhibitors = get_places_from_condition(tree.operand.__str__())
            inhibitors_nodes.update(current_inhibitors)
            LOGGER.debug("INHIBITORS found: " + str(current_inhibitors))

            path = [tree.operand.name]
            solutions = [path]
            return solutions
        except AttributeError:
            tree = tree.operand

    if isinstance(tree, SigIdentExpr):
        path = [tree.name]
        solutions = [path]
        return solutions



    lch = tree.left_h
    op  = tree.operator
    rch = tree.right_h
#    print('OZCJSH:', lch, op, rch, sep='\t\t')
    lpaths = rec(lch, inhibitors_nodes)
    rpaths = rec(rch, inhibitors_nodes)
#    print('VFUENK:', lpaths, rpaths)
    if op == 'or':  # or
#        ret = [*lpaths, *rpaths]
        ret = list(it.chain(lpaths, rpaths))
#        print('RET:', ret)
        return ret
    else:  # and
        assert op == 'and'
#        print(list(it.product(lpaths, rpaths)))
#        raw_input('test')

        ret = list(l + r for l, r in it.product(lpaths, rpaths))
#        print('RET:', ret)
        return ret


def get_places_from_condition(condition):
    """Parse condition string and return all places, regardless of operators.

    .. note:: This function is only used to get all nodes in a condition when
        we know they all are inhibitors nodes.

    :param: Condition string.
    :type: <str>
    :return: Set of places.
    :rtype: <set>
    """

    replacement = ['and', 'or', 'not', '(', ')']

    for chr in replacement:
        condition = condition.replace(chr, ' ')

    # Must be exempt of unauthorized chars
    return {elem for elem in condition.split(' ')
            if elem != ''}


def parse_condition(condition, all_nodes, inhibitors_nodes):
    """Return valid paths according the given logical formula and nodes.

    """

    LOGGER.debug("CONDITION: " + condition)
    # Error Reporter
    err = Reporter()
    tvi = TableVisitor(err)
    # Link the lexer to the model allows to avoid error in Reporter
    # like:  "-> dec -> Undeclared event or state"
    # In practice this is time consuming and useless for what we want to do
    #parser = MakeModelFromXmlFile(BIO_MOLDELS_DIR + "Whole NCI-PID database translated into CADBIOM formalism(and).bcx")
    #parser.get_model().accept(tvi)
    symb_tab = tvi.tab_symb
    # Get tree object from condition string
    cond_sexpr = compile_cond(condition, symb_tab, err)
    # Get all possible paths from the condition
    possible_paths = rec(cond_sexpr, inhibitors_nodes)

    # Prune possible paths according to:
    # - Inhibitor nodes that must be removed because they will never
    # be in the graph.
    # - All nodes in transitions (ori -> ext) because we know all transitions
    # in the graph, so we know which entities can be choosen to validate a path.
    # - All frontier places, that are known entities that can be in conditions
    # (not only in ori/ext) of transitions.
    # So: authorized nodes = frontier_places + transition_nodes - inhibitor nodes
    valid_paths = {tuple(path) for path in possible_paths
                   if (set(path) - inhibitors_nodes).issubset(all_nodes)}

    # Debugging only
    if LOGGER.getEffectiveLevel() == DEBUG:
        LOGGER.debug("INHIBIT NODES: " + str(inhibitors_nodes))
        LOGGER.debug("ALL NODES: " + str(all_nodes))
        LOGGER.debug("POSSIBLE PATHS: " + str(possible_paths))
        LOGGER.debug("VALID PATHS: " + str(valid_paths))

        for path in possible_paths:
            pruned_places = set(path) - inhibitors_nodes
            isinsubset = pruned_places.issubset(all_nodes)
            LOGGER.debug(
                "PRUNED PATH: {}, VALID: {}".format(
                    pruned_places,
                    isinsubset
                )
            )

    assert len(valid_paths) > 0, "No valid path for: " + str(condition)
    if len(valid_paths) > 1:
        LOGGER.warning("Multiple valid paths for: {}:\n{}".format(condition,
                                                                  valid_paths))
    return valid_paths


    # condition expressions contains only node ident
    icv = SigExpIdCollectVisitor()
    lst1 = cond_sexpr.accept(icv)
    print(cond_sexpr)
    print(type(cond_sexpr))
    print(dir(cond_sexpr))
    print("LISTE", lst1)
#    <class 'cadbiom.models.biosignal.sig_expr.SigSyncBinExpr'>
#    'accept', 'get_signals', 'get_ultimate_signals', 'is_bot', 'is_clock',
# 'is_const', 'is_const_false', 'is_ident', 'left_h', 'operator', 'right_h', 'test_equal']

    print(cond_sexpr.get_signals())
#    print(cond_sexpr.get_ultimate_signals())
    print("LEFT", cond_sexpr.left_h)
    print("OPERATOR", cond_sexpr.operator)
    print("RIGHT", cond_sexpr.right_h)


#    ret = treeToTab(cond_sexpr)
#    [set([('((formule', True)])]
#    print("treeToTab", ret)
#    print(type(ret))
#    print(dir(ret))


def parse_event(event):
    """Decompile logical formula in event's name.

    :param: Event string.
    :type: <event string>
    :return: A dict of events and their conditions.
    :rtype: <dict>
        keys: event's names; values: logical formula attached (condition)
    """

    def treeToExprDefaultsList(tree):
        if isinstance(tree, SigDefaultExpr):
            return treeToExprDefaultsList(tree.left_h) + \
                treeToExprDefaultsList(tree.right_h)

        else:
            # Here, some tree are from classes SigConstExpr or SigIdentExpr
            # Ex: for the clock "_h_5231":
            #    ... default (_h_5231)" => No condition for this event
            # Other examples:
            # _h_2018 _h_820 _h_4939 _h_5231 _h_3301 _h_4967 _h_2303 _h_3301
            return [tree]

    def filterSigExpression(expr):
        """
        .. note:: No SigConstExpr here => filtered in get_transitions()
            by checking null events (event="") in the model.
        """

        if isinstance(expr, SigWhenExpr):
            # right : SigSyncBinExpr (logical formula), BUT
            # sometimes SigConstExpr (just a True boolean) when clock is empty
            # Ex: "when ()"
            # So, we replace this boolean with an empty condition
            right = '' if isinstance(expr.right_h, SigConstExpr) \
                        else str(expr.right_h)

            return expr.left_h.name, right

        if isinstance(expr, SigIdentExpr):
            return expr.name, ''

        raise AssertionError("You should never have been there ! "
                             "Your expression type is not yet supported...")

#    def filterSigExpression(listOfExpr):
#        return [filterSigExpression(expr) for expr in listOfExpr]

    # Error Reporter
    err = Reporter()
    tvi = TableVisitor(err)
    symb_tab = tvi.tab_symb

    # Get tree object from event string
    event_sexpr = compile_event(event, symb_tab, True, err)[0]

    # Filter when events
    g = (filterSigExpression(expr) \
         for expr in treeToExprDefaultsList(event_sexpr))
    eventToCondStr = \
        {event_name: event_cond for event_name, event_cond in g}


    LOGGER.debug("Clocks from event parsing: " + str(eventToCondStr))

    return eventToCondStr


def build_graph(solution, steps, transitions):
    """Build a graph for the given solution.

        - Get & make all needed edges
        - Build graph

    :param arg1: Frontier places.
    :param arg2: List of steps (with events in each step).
    :param arg3: A dictionnary of events as keys, and transitions as values
        (see get_transitions()).
    :type arg1: <str>
    :type arg2: <list <list>>
    :type arg3: <dict <list <tuple <str>, <str>, <dict <str>: <str>>>>
    :return:
        - Networkx graph object.
        - Nodes corresponding to transitions with conditions.
        - All nodes in the model
        - Edges between transition node and nodes in condition
        - Normal transitions without condition
    :rtype: <networkx.classes.digraph.DiGraph>, <list>, <list>, <list>, <list>
    """

    def filter_transitions(step_event):
        """ Insert a transittion in a transition event if there is a condition.

        => Insert a node in a edge.
        => Link all nodes involved in the condition with this new node.

        :param: A list of events (transitions) (from a step in a solution).
            [('Ax', 'n1', {u'label': u'h00[]'}),]
        :type: <tuple>
        :return: Fill lists of edges:
            edges_with_cond: link to a transition node for
                transition with condition.
            transition_nodes: add new nodes corresponding to transitions with
                conditions.
            edges_in_cond: Add all edges to nodes linked to a transition
                via the condition (group of nodes involved in the path
                choosen by the solver).
            edges: transition untouched if there is no condition.
        :rtype: None
        """
        assert len(step_event) != 0 # Todo: useful ?

        inhibitors_nodes = set() # Inactivated nodes in paths of conditions
        input_places = {trans[0] for trans in step_event}

        # Color nodes
        # Since we explore all possible paths for each condition,
        # some edges are rewrited multiple times.
        # => included edges between origin <=> transition node
        # These edges must be grey while, edges between a node that is
        # only in a condition and a transition node must be green.
        # => notion of activator vs inhibitor vs normal input/output node
        def color_map(node):
            # print("color for:", node)
            if node in inhibitors_nodes: # Test first (see cond below)
                return 'red'
            if node in input_places: # some /all frontier places are in this set
                return 'grey'
            else:
                return 'green'


        for trans in step_event:
            attributes = trans[2]
            ori = trans[0]
            ext = trans[1]
            event = attributes['label'].split('[')[0]

            # If there is a condition formula associated to this clock
            if attributes['condition'] != '':

                # Add the transition as node
                transition_nodes.append(
                    (
                        event,
                        {
                            'label': attributes['label'],
                            'name': attributes['label'],
                            'color': 'blue',
                            'condition': attributes['condition'],
                        }
                    )
                )

                # Origin => transition node
                edges_with_cond.append(
                    (
                        ori, event,
                        {
                            'label': ori + '-' + event,
                        }
                    )
                )

                # Transition node => ext
                edges_with_cond.append(
                    (
                        event, ext,
                        {
                            'label': event + '-' + ext,
                        }
                    )
                )

                # Add all transitions to nodes linked via the condition
                valid_paths = parse_condition(
                    attributes['condition'],
                    all_nodes,
                    inhibitors_nodes
                )
                for i, path in enumerate(valid_paths):
                    for node in path:
                        edges_in_cond.append(
                            (
                                node, event,
                                {
                                    'label': '{} ({})'.format(
                                        event,
                                        i
                                    ),
#                                    'label': '{} [{}] ({})'.format(
#                                        event,
#                                        ', '.join(path),
#                                        i
#                                    ), #node + '-' + event,
                                    'color': color_map(node),
                                }
                            )
                        )
            else:
                # Normal edges
                edges.append(trans)


    # Get & make all needed edges ##############################################
    LOGGER.debug("BUILD GRAPH FOR SOL: " + str(solution))
    LOGGER.debug("STEPS: " + str(steps))
#
#    print(transitions['_h_2755'])
#    print(transitions['_h_4716'])
#    print(transitions['_h_2206'])
#    print(transitions['_h_3426'])
#    exit()

    frontier_places  = solution.split(' ')
    edges_with_cond  = list() # Edges between ori <-> transition node <-> ext
    edges_in_cond    = list() # Edges between transition node and nodes in condition
    transition_nodes = list() # Nodes inserted because of condition in transition
    edges            = list() # Normal transitions without condition

    # Get all nodes in all transitions (origin & ext)
    try:
        all_transitions = \
            (transitions[step_event] for step_event in it.chain(*steps))
        transitions_ori_ext = \
            (tuple((trans[0], trans[1])) for trans in it.chain(*all_transitions))
    except KeyError:
        print("/!\One event is not in the given model file... Please check it.")
        raise
    all_nodes = set(it.chain(*transitions_ori_ext)) | set(frontier_places)

    # Parse all conditions in transitions;
    # add nodes in conditions and transition nodes
    [filter_transitions(transitions[step_event])
        for step_event in it.chain(*steps)]

#    print("edges without cond", edges)
#    print("edges with cond", edges_with_cond)
#    print("transition nodes added", transition_nodes)
#    raw_input("PAUSE")

    # Make Graph ###############################################################
    G = nx.DiGraph()
    # Add all nodes (some frontier places are in this set)
    G.add_nodes_from(all_nodes, color='grey')
    # Add fontier places
    G.add_nodes_from(frontier_places, color='red')
    # Add all transition nodes
    G.add_nodes_from(transition_nodes, color='blue')

    # Node attribute ?
    # print(G.node['h1'])

    # Add all edges
    G.add_edges_from(edges)
    G.add_edges_from(edges_with_cond)
    G.add_edges_from(edges_in_cond)

    return G, transition_nodes, all_nodes, edges_in_cond, edges


def draw_graph(output_dir, solution, solution_index, G,
               transition_nodes, all_nodes,
               edges_in_cond, edges,
               matplotlib_export=False):
    """Draw graph with colors and export it to graphml format.

    .. note:: Legend:
        - red: frontier places (in solution variable),
        - white: middle edges,
        - blue: transition edges

    :param arg1: Output directory for graphml files.
    :param arg2: Solution string (mostly a set of frontier places).
    :param arg3: Index of the solution in the Cadbiom result file
        (used to distinguish exported filenames).
    :param arg4: Networkx graph object.
    :param arg5: Nodes corresponding to transitions with conditions.
        List of tuples: event, node
    :param arg6: All nodes in the model
    :param arg7: Edges between transition node and nodes in condition
    :param arg8: Normal transitions without condition
    :param arg9: If True, an image is exported by matplotlib (optional).
    :type arg1: <str>
    :type arg2: <str>
    :type arg3: <int> or <str>
    :type arg4: <networkx.classes.digraph.DiGraph>
    :type arg5: <list>
    :type arg6: <list>
    :type arg7: <list>
    :type arg7: <list>
    :type arg8: <boolean>
    """

    creation_date = dt.datetime.now().strftime("%H-%M-%S")
    # Save & show
    if not matplotlib_export:
        # Save the graph without matplotlib requirement
        # PS: inhibitors will still have not the attribute 'color' = 'white'
        nx.write_graphml(
            G,
            "{}{}_{}_{}.graphml".format(
                output_dir, creation_date, solution_index, solution[:75]
            )
        )
        return

    # Drawing ##################################################################
    # draw_circular(G, **kwargs) On a circle.
    # draw_random(G, **kwargs)   Uniformly at random in the unit square.
    # draw_spectral(G, **kwargs) Eigenvectors of the graph Laplacian.
    # draw_spring(G, **kwargs)   Fruchterman-Reingold force-directed algorithm.
    # draw_shell(G, **kwargs)    Concentric circles.
    # draw_graphviz(G[, prog])   Draw networkx graph with graphviz layout.

    # Get a list of transition nodes (without dictionnary of attributes)
    transition_nodes_names = [node[0] for node in transition_nodes]

    pos = nx.circular_layout(G)

    # Legend of conditions in transition nodes
    f = plt.figure(1)
    ax = f.add_subplot(1,1,1)
    text = '\n'.join(transition_nodes_names)
    ax.text(0, 0, text, style='italic', fontsize=10,
            bbox={'facecolor': 'white', 'alpha': 0.5, 'pad': 10})

    # Draw nodes:
    # - red: frontier places (in solution variable),
    # - white: middle edges,
    # - blue: transition edges
    frontier_places  = set(solution.split(' '))
    def color_map(node):
        # print("color for:", node)
        if node in frontier_places: # Test first (see cond below)
            return 'red'
        if node in unziped_transition_nodes:
            return 'blue'
        if node in all_nodes: # some /all frontier places are in this set
            return 'grey'
        else:
            return 'white'

    # Color nodes
    colors = [color_map(node) for node in G.nodes_iter()]
    nx.draw(G, pos=pos, with_labels=True,
            node_color=colors, node_size=1000, alpha=0.5,
            ax=ax)

    # Draw edges involved in transitions with conditions
    edges_colors = [edge[2]['color'] for edge in edges_in_cond]
    nx.draw_networkx_edges(G, pos, edgelist=edges_in_cond,
                           edge_color=edges_colors, width=2, alpha=0.5)

    # Draw labels for normal transitions (move pos to the end of the arrow)
    # ex: [('Ax', 'n1', {u'condition': u'', u'label': u'h00[]'}),]
    edges_labels = {(edge[0], edge[1]): edge[2]['label'] for edge in edges}
    nx.draw_networkx_edge_labels(G, pos, edges_labels, label_pos=0.3)

    # Save & show
    plt.legend()
    plt.savefig(
        output_dir + creation_date + '_' + solution[:75] + ".svg",
        format="svg")
    plt.show()


def process_solutions(output_dir, sol_steps, transitions):
    """Build a graph per solution"""

    for sol_index, (sol, steps) in enumerate(sol_steps):

        draw_graph(output_dir, sol, sol_index,
                   *build_graph(sol, steps, transitions)
        )


def test_main():
    """Test"""

    # chart_model.py
    # chart_xml.py
    parser = MakeModelFromXmlFile(BIO_MOLDELS_DIR + "mini_test_publi.bcx")
    print(type(parser.parser))
    print(dir(parser))
    print("HANDLER")
    print(dir(parser.handler))
    print(dir(parser.parser))
    print(dir(parser.get_model()))
    print("ICI")
#    print(parser.get_model().get_simple_node())
    print(parser.handler.node_dict)
    print(parser.handler.top_pile)
    print(parser.handler.pile_dict)
    print(parser.handler.transition.event)
    print(type(parser.handler.top_pile))

    transitions = dict()
    for transition in parser.handler.top_pile.transitions:
        # print(type(transition)) => list
        for trans in transition:
            # 'action', 'activated', 'clean', 'clock', 'condition', 'event',
            # 'ext', 'ext_coord', 'fact_ids', 'get_influencing_places',
            # 'get_key', 'is_me', 'macro_node', 'name', 'note', 'ori',
            # 'ori_coord', 'remove', 'search_mark', 'selected', 'set_action',
            # 'set_condition', 'set_event', 'set_name', 'set_note'

# {'name': '', 'clock': None, 'selected': False, 'activated': False,
# 'search_mark': False, 'note': '', 'ext': <cadbiom.models.guard_transitions.chart_model.CSimpleNode object at 0x7f391c7406d0>,
# 'ext_coord': 0.0, 'ori': <cadbiom.models.guard_transitions.chart_model.CSimpleNode object at 0x7f391c740650>,
# 'action': u'', 'macro_node': <cadbiom.models.guard_transitions.chart_model.CTopNode object at 0x7f391c7db490>,
# 'ori_coord': 0.0, 'event': u'h5', 'condition': u'', 'fact_ids': []}
            # print(dir(trans))
            print("NEW trans", trans.event)
#            print(trans.__dict__)
#            print(trans.name, trans.clock, trans.selected, trans.activated,
#                  trans.search_mark, trans.note, trans.ext, trans.ext_coord,
#                  trans.ori, trans.action, trans.macro_node, trans.ori_coord,
#                  trans.event, trans.condition, trans.fact_ids
#                  )

            transitions[trans.event] = trans.condition

#print("ORI", trans.ori.__dict__)
#{'name': 'n4', 'yloc': 0.906099768906, 'selected': False,
#'father': <cadbiom.models.guard_transitions.chart_model.CTopNode object at 0x7f09eed91490>,
#'xloc': 0.292715433748, 'search_mark': False, 'was_activated': False,
#'incoming_trans': [<cadbiom.models.guard_transitions.chart_model.CTransition object at 0x7f09eecf67d0>],
#'model': <cadbiom.models.guard_transitions.chart_model.ChartModel object at 0x7f09ef2cf3d0>,
#'outgoing_trans': [<cadbiom.models.guard_transitions.chart_model.CTransition object at 0x7f09eecf6850>],
#'activated': False, 'hloc': 1.0}
            print("ORI", trans.ori.name)
            try:
                print("ori INCO", [(tr.event, tr.condition) for tr in trans.ori.incoming_trans])
            except: pass
            try:
                print("ori OUTGO", [(tr.event, tr.condition) for tr in trans.ori.outgoing_trans])
            except: pass
            print("EXT", trans.ext.name)
            try:
                print("ext INCO", [(tr.event, tr.condition) for tr in trans.ext.incoming_trans])
            except: pass
            try:
                print("ext OUTGO", [(tr.event, tr.condition) for tr in trans.ext.outgoing_trans])
            except: pass
    print(transitions)



def sol_digging(sol_steps, transitions, conditions=True):
    """Convert all events for all solutions in a complete MAC file
    and write them in a separate file in the json format.

    This is a function to quickly search all transition attributes involved
    in a solution.

    :param arg1: List of steps involved in a solution. See load_solutions().
        A tuple of "frontier places" and a list of events in each step.
        ("Bx Ax", [[u'h2', u'h00'], [u'h3'], [u'h0', u'h1'], [u'hlast']])
    :param arg2: A dictionnary of events as keys, and transitions as values.
        Since many transitions can define an event, values are lists.
        Each transition is a tuple with: origin node, final node, attributes
        like label and condition.
        {u'h00': [('Ax', 'n1', {u'label': u'h00[]'}),]
        See get_transitions().
    :param arg3: (Optional) Integrate in the final file,
        the conditions for each transition.
    :type arg1: <list>
    :type arg2: <dict <list <tuple <str>, <str>, <dict <str>: <str>>>>
    :type arg3: <bool>
    :return: Return the JSON data for the given steps.
        [{
            "solution": "Ax Bx",
            "steps": [
                [{
                    "event": "_h_2",
                    "transitions": [{
                        "ext": "n3",
                        "ori": "Bx"
                    }]
                }],
            ]
        }]
    :rtype: <list>
    """

    def get_transition_def(step_event):
        """Dump each transition in the given event to a list of dictionnaries.

        .. note:: ori="JUN_nucl_gene" ext="JUN_nucl" event="_h_391"
        :return: A list of dictionaries
            (1 dict for 1 transition in the given event)
        :rtype: <list <dict>>
        """

        # Many transitions per event (ex: complex dissociation)
        decomp_transitions = list()
        for trans in step_event:

            decomp_transition = {
                #"event": trans[2]['label'].split('[')[0],
                "ori": trans[0],
                "ext": trans[1],
            }
            # If condition boolean is set (by default),
            # we add the event's transition to the json data.
            if conditions:
                decomp_transition["condition"] = trans[2]['condition']

            decomp_transitions.append(decomp_transition)

        return decomp_transitions


    # sol_steps structure:
    # ("Bx Ax", [[u'h2', u'h00'], [u'h3'], [u'h0', u'h1'], [u'hlast']])
    decomp_solutions = list()
    for sol, steps in sol_steps:
        # Decompile steps in each solution
        decomp_steps = list()
        for step in steps:
            # Decompile events in each step
            decomp_events = list()
            for event in step:
                # Decompile transitions in each event
                decomp_event = dict()
                # Get transitions for the given event
                # Structure of transitions:
                # {u'h00': [('Ax', 'n1', {u'label': u'h00[]'}),]
                step_event = transitions[event]
                if len(step_event) == 0:
                    decomp_event['event'] = "ERROR, no transition"
                else:
                    # Get list of transitions
                    decomp_event['event'] = event
                    decomp_event['transitions'] = get_transition_def(step_event)
                # Add event and its transitions
                decomp_events.append(decomp_event)
            # Add step and its events
            decomp_steps.append(decomp_events)
        # Add solution and its steps
        solution = {
            "solution": sol,
            "steps": decomp_steps,
        }
        decomp_solutions.append(solution)

    # Debugging only
    if LOGGER.getEffectiveLevel() == DEBUG:
        LOGGER.debug(json.dumps(decomp_solutions, sort_keys=True, indent=4))

    return decomp_solutions


def sol_digging_main(output_dir, model_file, solution_path, conditions=True):
    """Entry point for sol_digging

    .. note:: This functions tests if the solution_path is a directory
        or just a file.
    """

    def write_json_file(decompiled_filename, decomp_solutions):
        # Write file
        with open(decompiled_filename, 'w') as f_d:
            json.dump(decomp_solutions, f_d, sort_keys=True, indent=4)

    # Get transitions from the model
    model_transitions = get_transitions(model_file)

    if os.path.isfile(solution_path):
        # The given path is a solution file
        # Add _decomp to the solution filename
        filename = os.path.basename(os.path.splitext(solution_path)[0])
        decompiled_filename = output_dir + filename + '_decomp.txt'
        decomp_solutions = sol_digging(
            load_solutions(solution_path),
            model_transitions,
            conditions=conditions,
        )
        write_json_file(decompiled_filename, decomp_solutions)

    elif os.path.isdir(solution_path):
        # The given path is a directory
        solution_path = solution_path if solution_path[-1] == '/' \
            else solution_path + '/'

        # Decompilation of all files in the directory
        for file_number, solution_file in \
            enumerate(glob.glob(solution_path + '*cam_complete.txt'), 1):

            # Add _decomp to the solution filename
            filename = os.path.basename(os.path.splitext(solution_file)[0])
            decompiled_filename = output_dir + filename + '_decomp.txt'
            decomp_solutions = sol_digging(
                load_solutions(solution_file),
                model_transitions,
                conditions=conditions,
            )
            write_json_file(decompiled_filename, decomp_solutions)

        LOGGER.info("Files processed: " + str(file_number))


def parse_trajectories_main(output_dir, model_file, solution_file):
    """Entry point for parse_trajectories"""

    process_solutions(
        output_dir,
        load_solutions(solution_file),
        get_transitions(model_file)
    )


def graph_isomorph_test(model_file_1, model_file_2, output_dir='graphs/',
                        make_graphs=False, make_json=False):
    """Entry point for model consistency checking.

    This functions checks if the 2 given models have the same topology,
    nodes & edges attributes/roles.

    .. note:: cf graphmatcher
        https://networkx.github.io/documentation/development/reference/generated/networkx.algorithms.isomorphism.categorical_edge_match.html

    :param arg1: File for the model 1.
    :param arg2: File for the model 2.
    :param arg3: Output path.
    :param arg4: If True, make a graphml file in output path.
    :param arg5: If True, make a json dump of results in output path.
    :type arg1: <str>
    :type arg2: <str>
    :type arg3: <str>
    :type arg4: <boolean>
    :type arg5: <boolean>
    :return: Dictionary with the results of tests.
        keys: 'topology', 'nodes', 'edges'; values: booleans
    :rtype: <dict <str>: <boolean>>
    """

    import networkx.algorithms.isomorphism as iso

    # Load transitions in the models
    # Transitions structure format:
    # {u'h00': [('Ax', 'n1', {u'label': u'h00[]'}),]
    transitions_1, all_places_1 = get_transitions(model_file_1, all_places=True)
    transitions_2, all_places_2 = get_transitions(model_file_2, all_places=True)

    # Get all frontier places in the models
    # (places that are never in output position in all transitions)
    # EDIT: why we use all_places from the model instead of
    # (input_places - output_places) to get frontier places ?
    # Because some nodes are only in conditions and not in transitions.
    # If we don't do that, these nodes are missing when we compute
    # valid paths from conditions.
    front_places_1 = " ".join(get_frontier_places(transitions_1, all_places_1))
    front_places_2 = " ".join(get_frontier_places(transitions_2, all_places_2))
    LOGGER.debug("Frontier places 1: " + str(front_places_1))
    LOGGER.debug("Frontier places 2: " + str(front_places_2))

    # Build graphs & get networkx object
    # We give all events in the model as a list of steps
    # So we simulate a cadbiom solution (with all events in the model).
    res_1 = build_graph(front_places_1, [transitions_1.keys()], transitions_1)
    G1 = res_1[0]
    res_2 = build_graph(front_places_2, [transitions_2.keys()], transitions_2)
    G2 = res_2[0]

    # Checking
    nm = iso.categorical_node_match('color', 'grey')
    em = iso.categorical_edge_match('color', '')

    check_state = \
    {
        'topology': nx.is_isomorphic(G1, G2),
        'nodes': nx.is_isomorphic(G1, G2, node_match=nm),
        'edges': nx.is_isomorphic(G1, G2, edge_match=em),
    }

    LOGGER.info("Topology checking: " + str(check_state['topology']))
    LOGGER.info("Nodes checking: " + str(check_state['nodes']))
    LOGGER.info("Edges checking: " + str(check_state['edges']))

    # Draw graph
    if make_graphs:
        draw_graph(output_dir, front_places_1, "first", *res_1)
        draw_graph(output_dir, front_places_2, "second", *res_2)

    # Export to json file
    if make_json:
        with open(output_dir + "comp_results.json", 'w') as fd:
            fd.write(json.dumps(check_state, sort_keys=True, indent=4) + '\n')

    return check_state


def graph_infos(model_file, output_dir='graphs/',
                make_graph=True, make_json=True, advanced_measures=False):
    """Entry point for model stats.

    :param arg1: File for the model.
    :param arg2: Output path.
    :param arg3: If True, make a graphml file in output path.
    :param arg4: If True, make a json dump of results in output path.
    :type arg1: <str>
    :type arg2: <str>
    :type arg3: <boolean>
    :type arg4: <boolean>
    :return: Dictionary with the results of measures on the given graph.
        keys: measure's name; values: measure's value
    :rtype: <dict>
    """

    # Load transitions in the model
    # Transitions structure format:
    # {u'h00': [('Ax', 'n1', {u'label': u'h00[]'}),]
    transitions_1, all_places_1 = get_transitions(model_file, all_places=True)

    # Get all frontier places in the models
    # (places that are never in output position in all transitions)
    # EDIT: why we use all_places from the model instead of
    # (input_places - output_places) to get frontier places ?
    # Because some nodes are only in conditions and not in transitions.
    # If we don't do that, these nodes are missing when we compute
    # valid paths from conditions.
    front_places = get_frontier_places(transitions_1, all_places_1)
    front_places_str = " ".join(front_places)
    LOGGER.debug("Frontier places 1: " + front_places_str)

    # Build graphs & get networkx object
    # We give all events in the model as a list of steps
    # So we simulate a cadbiom solution (with all events in the model).
    res_1 = build_graph(front_places_str, [transitions_1.keys()], transitions_1)
    G = res_1[0]

    infos = {
        'model': model_file,
        'nodes': len(G.nodes()),
        'edges': len(G.edges()),
        'transitions:': len(transitions_1),
        'places': len(all_places_1),
        'frontier places': len(front_places),
    }

    if advanced_measures:
        infos.update({
            'degree': nx.degree_centrality(G),
            'in_degree': nx.in_degree_centrality(G),
            'out_degree': nx.out_degree_centrality(G),
            'betweenness': nx.betweenness_centrality(G),
            'closeness': nx.closeness_centrality(G),
        })

    # Draw graph
    if make_graph:
        draw_graph(output_dir, front_places_str, 'stat', *res_1)

    # Export to json file
    if make_json:
        with open(output_dir + "infos_results.json", 'w') as fd:
            fd.write(json.dumps(infos, sort_keys=True, indent=4) + '\n')

    LOGGER.info(str(infos))
    return infos


if __name__ == "__main__":

    LOG_DIR = "./logs/"
    BIO_MOLDELS_DIR = "./bio_models/"

    graph_isomorph_test(BIO_MOLDELS_DIR + "mini_test_publi.bcx",
                        BIO_MOLDELS_DIR + "mini_test_publi_mod.bcx")
    exit()

    process_solutions(load_solutions(LOG_DIR + "../run/Whole NCI-PID database translated into CADBIOM formalism(and)_SRP9_cam_complete.txt"),
                get_transitions(BIO_MOLDELS_DIR + "Whole NCI-PID database translated into CADBIOM formalism(and).bcx"))
    exit()


    #sort_solutions("/media/DATA/Projets/dyliss_tgf/cadbiom/run/Whole NCI-PID database translated into CADBIOM formalism(and)_SRP9_cam_complete.txt")
    #sort_solutions("/media/DATA/Projets/dyliss_tgf/cadbiom/run/Whole NCI-PID database translated into CADBIOM formalism(and)_SRP9_cam.txt")
    #sort_solutions("/media/DATA/Projets/dyliss_tgf/cadbiom/run/pid_and_clock_no_perm_p21corrected_start_SRP9_complete.txt")
    #exit()

#    cond = "((((not(ATF2_JUND_macroH2A_nucl))or(Fra1_JUND_active_nucl))or(Fra1_JUN_active_nucl))or(TCF4_betacatenin_active_nucl))or(JUN_FOS_active_nucl)"
#    ret = parse_condition(cond)
#    print(ret)
#    exit()

#    sol_steps = load_solutions(LOG_DIR + "../run/pid_and_clock_SRP9_cam_complete_o.txt")
#    g = [sol_step for sol_step in sol_steps]
#    print(g)
#    exit()

    sol_digging(load_solutions(LOG_DIR + "../run/pid_and_clock_no_perm_p21corrected_start_SRP9_complete.txt"),
             get_transitions(BIO_MOLDELS_DIR + "Whole NCI-PID database translated into CADBIOM formalism(and).bcx"))
    exit()

    sol_digging(load_solutions(LOG_DIR + "sols_new_solver.txt"),
             get_transitions(BIO_MOLDELS_DIR + "mini_test_publi.bcx"))
    exit()
    process_solutions(load_solutions(LOG_DIR + "sols_new_solver.txt"),
                      get_transitions(BIO_MOLDELS_DIR + "mini_test_publi.bcx"))
    exit()

#    build_graph(load_solutions(LOG_DIR + "../run/pid_and_clock_SRP9_cam_complete_o.txt"),
#                get_transitions(BIO_MOLDELS_DIR + "pid_and_clock.bcx"))
    process_solutions(load_solutions(LOG_DIR + "../run/pid_and_clock_SRP9_cam_complete.txt"),
                get_transitions(BIO_MOLDELS_DIR + "pid_and_clock.bcx"))
